﻿/// Модуль предоставляет средства для разработки свёрточных, полносвёрточных и полносвязных нейронных сетей.
/// (C) PABCSoft, 2018
unit ConvNetABC;
{$Reference 'System.dll'}
{$Reference 'mscorlib.dll'}
{$Reference 'System.IO.dll'}
{$Reference 'System.Drawing.dll'}
{$Reference 'System.Threading.dll'}
{$Reference 'System.Collections.dll'}
{$GenDoc True}
interface

Uses System;
Uses System.IO;
Uses Converters;
Uses System.Drawing;
Uses System.Threading;
Uses System.Threading.Tasks;
Uses System.Collections.Generic;

type
  ///<summary>Представляет результат операции поиска минимума и максимума тензора.</summary>
  maxminResult = Record
    ///<summary>Индекс максимального элемента.</summary>
    public maxi: Single;
    ///<summary>Значение максимального элемента.</summary>
    public maxv: Single;
    ///<summary>Индекс минимального элемента.</summary>
    public mini: Single;
    ///<summary>Значение минимального элемента.</summary>
    public minv: Single;
    ///<summary>Разброс значений элементов массива.</summary>
    public dv: Single;
  End;

type
  ///<summary>Предоставляет базовые операции, такие как генераторы случайных чисел и поиск максимума/минимума тензора.</summary>
  Util = {static} class
    public class Return_V: Boolean;
    public class V_Val: Single;
    private class random: Random := new Random(3);
    public class function GaussRandom(): Single;
    begin
      if (Return_V)
      Then
      begin
        Return_V := false;
        Result := V_Val;
        exit;
      end;
      var u := 2 * random.NextDouble() - 1;
      var v := 2 * random.NextDouble() - 1;
      var r := (u * u) + (v * v);
      if ((r = 0) or (r > 1))
      Then
      begin
        Result := GaussRandom();
        exit;
      end;
      var c := Math.Sqrt(-2 * Math.Log(r) / r);
      V_Val := Double(v * c);
      Return_V := true;
      Result := Double(u * c);
      exit;
    end;
    
    public class function RandF(a: Single; b: Single): Single;
    begin
      Result := Single(random.NextDouble() * (b - a) + a);
      exit;
    end;
    
    public class function Randi(a: Single; b: Single): Single;
    begin
      Result := Single(Math.Floor(random.NextDouble() * (b - a) + a));
      exit;
    end;
    
    public class function Randn(mu: Single; std: Single): Single;
    begin
      Result := mu + GaussRandom() * std;
    end;
    
    public class function ArrContains(arr: array of Single; elt: Single): Boolean;
    begin
      Result := arr.Contains(elt);
      exit;
    end;
    
    public class function ArrUnique(arr: array of Single): array of Single;
    begin
      var d := new List<Single>();
      for Var i := 0 to arr.Length - 1 do
      begin
        if (not d.Contains(arr[i]))
        Then
        begin
          d.Add(arr[i]);
        end;
      end;
      Result := d.ToArray();
    end;
    
    public class function maxmin(w: array of Single): maxminResult;
    begin
      var maxv := w[0];
      var minv := w[0];
      var maxi := 0;
      var mini := 0;
      var n := w.Length;
      for Var i := 0 to n - 1 do
      begin
        if (w[i] > maxv)
        Then
        begin
          maxv := w[i];
          maxi := i;
        end;
        if (w[i] < minv)
        Then
        begin
          minv := w[i];
          mini := i;
        end;
      end;
      Result := new maxminResult();
      Result.maxi := maxi;
      Result.maxv := maxv;
      Result.mini := mini;
      Result.minv := minv;
      Result.dv := maxv - minv;
      exit;
    end;
    
    public class function RandPerm(n: Int32): array of Single;
    begin
      var i := n;
      var j := 0;
      var temp: Single := 0.0;
      var &array := new Single[n];
      for Var q := 0 to n - 1 do
      begin
        &array[q] := q;
      end;
      while (i > 0) do
      begin
        i -= 1;
        j := Int32(Math.Floor( random.NextDouble() * (i + 1)));
        temp := &array[i];
        &array[i] := &array[j];
        &array[j] := temp;
      end;
      Result := &array;
      exit;
    end;
    
    public class function WeightedSample(lst: array of Single; probs: array of Single): Single;
    begin
      var p := RandF(0, 1);
      var cumprob := 0.0;
      for Var k := 0 to lst.Length - 1 do
      begin
        cumprob += probs[k];
        if (p < cumprob)
        Then
        begin
          Result := lst[k];
          exit;
        end;
      end;
      Result := -1;
      exit;
    end;
  End;

type
  DataToSave = class
    public Data: array of array of Single;
    public Labels: array of Single;
  End;

type
  Vol = class
    public Width: Int32;
    public Height: Int32;
    public Depth: Int32;
    public W: array of Single;
    public DW: array of Single;
    public step_cache: array of Single;
    public constructor Create(sx: array of Single);
    begin
      self.Width := 1;
      self.Height := 1;
      self.Depth := sx.Length;
      self.W := new Single[self.Depth];
      self.DW := new Single[self.Depth];
      self.step_cache := new Single[self.Depth];
      for Var i := 0 to self.Depth - 1 do
      begin
        self.W[i] := sx[i];
      end;
    end;
    
    public constructor Create(sx: Int32; sy: Int32; depth: Int32; normal: Boolean := false);
    begin
      self.Width := sx;
      self.Height := sy;
      self.Depth := depth;
      var n := sx * sy * depth;
      self.W := new Single[n];
      self.DW := new Single[n];
      self.step_cache := new Single[n];
      var scale := Single(Math.Sqrt(1.0 / (sx * sy * depth)));
      if (normal)
      Then
      begin
        scale := 0.08;
      end;
      for Var i := 0 to n - 1 do
      begin
        self.W[i] := Util.Randn(0.0, scale);  
      end;
    end;
    
    public constructor(sx: Int32; sy: Int32; depth: Int32; c: Single);
    begin
      self.Width := sx;
      self.Height := sy;
      self.Depth := depth;
      var n := sx * sy * depth;
      self.W := new Single[n];
      self.DW := new Single[n];
      self.step_cache := new Single[n]; 
      for Var i := 0 to n - 1 do
      begin
        self.W[i] := c;
      end;
    end;
    
    public function Get(x: Int32; y: Int32; d: Int32): Single;
    begin
      var ix := ((self.Width * y) + x) * self.Depth + d;
      Result := self.W[ix];
    end;
    
    public procedure &Set(x: Int32; y: Int32; d: Int32; v: Single);
    begin
      var ix := ((self.Width * y) + x) * self.Depth + d;
      self.W[ix] := v;
    end;
    
    public procedure Add(x: Int32; y: Int32; d: Int32; v: Single);
    begin
      var ix := ((self.Width * y) + x) * self.Depth + d;
      self.W[ix] += v;
    end;
    
    public function Get_Grad(x: Int32; y: Int32; d: Int32): Single;
    begin
      var ix := ((self.Width * y) + x) * self.Depth + d;
      Result := self.DW[ix];
    end;
    
    public procedure Set_Grad(x: Int32; y: Int32; d: Int32; v: Single);
    begin
      var ix := ((self.Width * y) + x) * self.Depth + d;
      self.DW[ix] := v;
    end;
    
    public procedure Add_Grad(x: Int32; y: Int32; d: Int32; v: Single);
    begin
      var ix := ((self.Width * y) + x) * self.Depth + d;
      self.DW[ix] += v;
    end;
    
    public function CloneAndZero(): Vol;
    begin
      Result := new Vol(self.Width, self.Height, self.Depth, 0.0);
    end;
    
    public function Clone(): Vol;
    begin
      var v := new Vol(self.Width, self.Height, self.Depth, 0.0);
      var n := self.W.Length;
      for Var i := 0 to n - 1 do
      begin
        v.W[i] := self.W[i];
      end;
      Result := v;
    end;
    
    public procedure AddFrom(V: Vol);
    begin
      for Var k := 0 to self.W.Length - 1 do
      begin
        self.W[k] += V.W[k];
      end;
    end;
    
    public procedure AddFromScaled(V: Vol; a: Single);
    begin
      for Var k := 0 to self.W.Length - 1 do
      begin
        self.W[k] += a * V.W[k];
      end;
    end;
    
    public procedure SetConst(a: Single);
    begin
      for Var k := 0 to self.W.Length - 1 do
      begin
        self.W[k] := a;
      end;
    end;
    
    public class function Augment(V: Vol; corp: Int32; dx: Int32; dy: Int32; fliplr: Boolean): Vol;
    begin
      var W: Vol;
      if ((corp <> V.Width) or (dx <> 0) or (dy <> 0))
      Then
      begin
        W := new Vol(corp, corp, V.Depth, 0.0);
        for Var x := 0 to corp - 1 do
        begin
          for Var y := 0 to corp - 1 do
          begin
            if ((x + dx < 0) or (x + dx >= V.Width) or (y + dy < 0) or (y + dy >= V.Height))
              Then continue;
            for Var d := 0 to V.Depth - 1 do
            begin
              W.Set(x, y, d, V.Get(x + dx, y + dy, d));
            end;
          end;
        end;
      end
      else
      begin
        W := V;
      end;
      if (fliplr)
      Then
      begin
        var W2 := W.CloneAndZero();
        for Var x := 0 to W.Width - 1 do
        begin
          for Var y := 0 to W.Height - 1 do
          begin
            for Var d := 0 to W.Depth - 1 do
            begin
              W2.Set(x, y, d, W.Get(W.Width - x - 1, y, d));
            end;
          end;
        end;
        W := W2;
      end;
      Result := W;
    end;
    
    public class function Augment(V: Vol; corp: Int32): Vol;
    begin
      var dx: Int32;
      var dy: Int32;
      var fliplr: Boolean := false;
      dx := Int32(Util.Randi(0, V.Width - corp));
      dy := Int32(Util.Randi(0, V.Height - corp)); 
      var W: Vol;
      if ((corp <> V.Width) or (dx <> 0) or (dy <> 0))
      Then
      begin
        W := new Vol(corp, corp, V.Depth, 0.0);
        for Var x := 0 to corp - 1 do
        begin
          for Var y := 0 to corp - 1 do
          begin
            if ((x + dx < 0) or (x + dx >= V.Width) or (y + dy < 0) or (y + dy >= V.Height))
              Then continue;
            for Var d := 0 to V.Depth - 1 do
            begin
              W.Set(x, y, d, V.Get(x + dx, y + dy, d));
            end;
          end;
        end;
      end
      else
      begin
        W := V;
      end;
      if (fliplr)
      Then
      begin
        var W2 := W.CloneAndZero();
        for Var x := 0 to W.Width - 1 do
        begin
          for Var y := 0 to W.Height - 1 do
          begin
            for Var d := 0 to W.Depth - 1 do
            begin
              W2.Set(x, y, d, W.Get(W.Width - x - 1, y, d));
            end;
          end;
        end;
        W := W2;
      end;
      Result := W;
    end;
    
    public class function Image_To_Vol(img: array of Single; width: Int32; hight: Int32; convert_grayscale: Boolean): Vol;
    begin
      var p := img;
      var w := width;
      var h := hight;
      var pv: array of Single := new Single[img.Length];
      for var i := 0 to p.Length - 1 do
      begin
        pv[i] := (p[i] / 255.0 - 0.5);
      end;
      var x := new Vol(w, h, 4, 0.0);
      x.W := pv;
      if (convert_grayscale)
      Then
      begin
        var x1 := new Vol(w, h, 1, 0.0);
        for Var i := 0 to w - 1 do
        begin
          for Var j := 0 to h - 1 do
          begin
            x1.Set(i, j, 0, x.Get(i, j, 0));
          end;
        end;
        x := x1;
      end;
      Result := x;
    end;
  End;

type
  ParamsAndGrads = class
    public &Params: array of Single;
    public Grads: array of Single;
    public l1_decay_mul: Single;
    public l2_decay_mul: Single;
  End;

type
  ///<summary>Интерфейс слоя нейросети.</summary>
  ILayer = Interface
    ///<summary>Метод обратного распространения(ошибки) через слой.</summary>
    procedure Backward();
    ///<summary>Метод прямого распространения через слой.</summary>
    ///<param name="V">Входной тензор.</param>
    ///<param name="is_training">Значение, указывающее на то, является ли распространение обучающим.</param>
    function &Forward(V: Vol; is_training: Boolean): Vol;
    function getParamsAndGrads(): List<ParamsAndGrads>;
    property Output: Vol read write;
    property InputDepth: Int32 read write;
    property InputWidth: Int32 read write;
    property InputHeight: Int32 read write;
    property OutputDepth: Int32 read write;
    property OutputWidth: Int32 read write;
    property OutputHeight: Int32 read write;
  End;

type
  ILossLayer = Interface
    function Backward(y: array of Single): Single;
    function Backward(y: Int32): Single;
    function &Forward(V: Vol; is_training: Boolean): Vol;
    function getParamsAndGrads(): List<ParamsAndGrads>;
    property Output: Vol read write;
  End;

type
  ClassOutput = Record
    public dim: Int32;
    public val: Single;
  End;

type
  RegressionLayer = class(ILossLayer)
    public Out_Depth: Int32;
    public out_sx: Int32;
    public out_sy: Int32;
    public num_inputs: Int32;
    public input: Vol;
    private __Output: Vol;
    public property Output: Vol read __Output write __Output;
    public constructor Create(inputDepth: Int32; inputWidth: Int32; inputHeight: Int32);
    begin
      self.num_inputs := inputWidth * inputHeight * inputDepth;
      self.Out_Depth := self.num_inputs;
      self.out_sx := 1;
      self.out_sy := 1;
    end;
    
    public function &Forward(V: Vol; is_training: Boolean): Vol;
    begin
      self.input := V;
      self.Output := V;
      Result := V;
    end;
    
    public function Backward(y: array of Single): Single;
    begin
      var x := self.input;
      x.DW := new Single[x.W.Length];
      var loss := 0.0;
      for var i := 0 to self.Out_Depth - 1 do
      begin
        var dy := x.W[i] - y[i];
        x.DW[i] := dy;
        loss += 0.5 * dy * dy;
      end;
      Result := loss;
    end;
    
    public function Backward(y: Int32): Single;
    begin
      var x := self.input;
      x.DW := new Single[x.W.Length];
      var loss := 0.0;
      var dy := x.W[0] - y;
      x.DW[0] := dy;
      loss += 0.5 * dy * dy;
      Result := loss;
    end;
    
    public function Backward(y: ClassOutput): Single;
    begin
      var x := self.input;
      x.DW := new Single[x.W.Length];
      var loss := 0.0;
      var i := y.dim;
      var yi := y.val;
      var dy := x.W[i] - yi;
      x.DW[i] := dy;
      loss += 0.5 * dy * dy;
      Result := loss;
    end;
    
    public function getParamsAndGrads(): List<ParamsAndGrads>;
    begin
      var response := new List<ParamsAndGrads>();
      Result := response;
    end;
  End;

type
  SoftmaxLayer = class(ILossLayer)
    public Out_Depth: Int32;
    public out_sx: Int32;
    public out_sy: Int32;
    public SX: Int32;
    public SY: Int32;
    public num_inputs: Int32;
    public es: array of Single;
    public input: Vol;
    private __Output: Vol;
    public property Output: Vol read __Output write __Output;
    public constructor Create(inputDepth: Int32; inputWidth: Int32; inputHeight: Int32);
    begin
      self.num_inputs := inputWidth * inputHeight * inputDepth;
      self.Out_Depth := self.num_inputs;
      self.out_sx := 1;
      self.out_sy := 1;
    end;
    
    public function &Forward(V: Vol; is_training: Boolean): Vol;
    begin
      self.input := V;
      var A := new Vol(1, 1, self.Out_Depth, 0.0);
      var ass := V.W;
      var amax := V.W[0];
      for var i := 1 to ass.Length - 1 do
      begin
        if (ass[i] > amax)
          Then amax := ass[i];
      end;
      var es := new Single[self.Out_Depth];
      var esum := 0.0;
      for var i := 0 to self.Out_Depth - 1 do
      begin
        var e := Single(Math.Exp(ass[i] - amax));
        esum += e;
        es[i] := e;
      end;
      for var i := 0 to self.Out_Depth - 1 do
      begin
        es[i] /= Single(esum);
        A.W[i] := es[i];
      end;
      self.es := es;
      self.Output := A;
      Result := self.Output;
    end;
    
    public function Backward(y: Int32): Single;
    begin
      var x := self.input;
      x.DW := new Single[x.W.Length];
      for var i := 0 to self.Out_Depth - 1 do
      begin
        var indicator := (i = y) ? (1.0) : (0.0);
        var mul := Single(-(indicator - self.es[i]));
        x.DW[i] := mul;
      end;
      Result := Single(-Math.Log(self.es[y]));
    end;
    
    public function getParamsAndGrads(): List<ParamsAndGrads>;
    begin
      var response := new List<ParamsAndGrads>();
      Result := response;
    end;
    
    public function Backward(y: array of Single): Single;
    begin
      raise new NotImplementedException();
    end;
  End;

type
  SVMLayer = class(ILossLayer)
    public Out_Depth: Int32;
    public out_sx: Int32;
    public out_sy: Int32;
    public num_inputs: Int32;
    public input: Vol;
    private __Output: Vol;
    public property Output: Vol read __Output write __Output;
    public constructor Create(inputDepth: Int32; inputWidth: Int32; inputHeight: Int32);
    begin
      self.num_inputs := inputWidth * inputHeight * inputDepth;
      self.Out_Depth := self.num_inputs;
      self.out_sx := 1;
      self.out_sy := 1;
    end;
    
    public function &Forward(V: Vol; is_training: Boolean): Vol;
    begin
      self.input := V;
      self.Output := V;
      Result := V;
    end;
    
    public function Backward(y: Int32): Single;
    begin
      var x := self.input;
      x.DW := new Single[x.W.Length];
      var yscore := x.W[y];
      var margin := 1.0;
      var loss := 0.0;
      for var i := 0 to self.Out_Depth - 1 do
      begin
        if (y = i)
          Then continue;
        var ydiff := -yscore + x.W[i] + margin;
        if (ydiff > 0)
        Then
        begin
          x.DW[i] += 1;
          x.DW[y] -= 1;
          loss += ydiff;
        end;
      end;
      Result := Single(loss);
    end;
    
    public function getParamsAndGrads(): List<ParamsAndGrads>;
    begin
      var response := new List<ParamsAndGrads>();
      Result := response;
    end;
    
    public function Backward(y: array of Single): Single;
    begin
      raise new NotImplementedException();
    end;
  End;

type
  FullyConnLayer = class(ILayer)
    private __OutputDepth: Int32;
    private __OutputHeight: Int32;
    private __OutputWidth: Int32;
    private __InputDepth: Int32;
    private __InputHeight: Int32;
    private __InputWidth: Int32;
    public property OutputDepth: Int32 read __OutputDepth write __OutputDepth;
    public L1_Decay_Mul: Single;
    public L2_Decay_Mul: Single;
    public property OutputWidth: Int32 read __OutputWidth write __OutputWidth;
    public property OutputHeight: Int32 read __OutputHeight write __OutputHeight;
    public num_inputs: Int32;
    public Biases: Vol;
    public Filters: array of Vol; 
    public input: Vol;
    private __Output: Vol;
    public property Output: Vol read __Output write __Output;
    public constructor Create(num_neurons: Int32; inputDepth: Int32; inputWidth: Int32; inputHeight: Int32);
    begin
      Init(num_neurons, inputDepth, inputWidth, inputHeight, 0.0, 1.0, 0.0);
    end;
    
    public constructor Create(num_neurons: Int32; inputDepth: Int32; inputWidth: Int32; inputHeight: Int32; l1_decay: Single; l2_decay: Single; bais_pref: Single);
    begin
      Init(num_neurons, inputDepth, inputWidth, inputHeight, l1_decay, l2_decay, bais_pref);
    end;
    
    private procedure Init(num_neurons: Int32; inputDepth: Int32; inputWidth: Int32; inputHeight: Int32; l1_decay: Single; l2_decay: Single; bais_pref: Single);
    begin
      self.OutputDepth := num_neurons;
      self.InputHeight := inputHeight;
      self.InputWidth := inputWidth;
      self.InputDepth := inputDepth;
      self.L1_Decay_Mul := l1_decay;
      self.L2_Decay_Mul := l2_decay;
      self.num_inputs := inputWidth * inputHeight * inputDepth;
      self.OutputWidth := 1;
      self.OutputHeight := 1;
      var bais := bais_pref;
      self.Filters := new Vol[self.OutputDepth];
      for Var i := 0 to self.OutputDepth - 1 do
      begin
        self.Filters[i] := (new Vol(1, 1, self.num_inputs));
      end;
      self.Biases := new Vol(1, 1, self.OutputDepth, bais);
    end;
    
    public function &Forward(V: Vol; is_training: Boolean): Vol;
    begin
      self.input := V;
      var A := new Vol(1, 1, self.OutputDepth, 0.0);
      var Vw := V.W;
      for Var i := 0 to self.OutputDepth - 1 do
      begin
        var _a := 0.0;
        var wi := self.Filters[i].W;
        for var d := 0 to self.num_inputs - 1 do
        begin
          _a += Vw[d] * wi[d];
        end;
        _a += self.Biases.W[i];
        A.W[i] := _a;
      end;
      self.Output := A;
      Result := self.Output;
    end;
    
    public procedure Backward();
    begin
      var V := self.input;
      V.DW := new Single[V.W.Length];
      for var i := 0 to self.OutputDepth - 1 do
      begin
        var tfi := self.Filters[i];
        var chain_grad := self.Output.DW[i];
        for var d := 0 to self.num_inputs - 1 do
        begin
          V.DW[d] += tfi.W[d] * chain_grad;
          tfi.DW[d] += V.W[d] * chain_grad;
        end;
        self.Biases.DW[i] += chain_grad;
      end;
    end;
    
    public function getParamsAndGrads(): List<ParamsAndGrads>;
    begin
      var response := new List<ParamsAndGrads>();
      for var i := 0 to self.OutputDepth - 1 do
      begin
        var pg := new ParamsAndGrads();
        pg.Params := self.Filters[i].W;
        pg.Grads := self.Filters[i].DW;
        pg.l2_decay_mul := self.L2_Decay_Mul;
        pg.l1_decay_mul := self.L1_Decay_Mul;
        response.Add(pg);
      end;
      var pg := new ParamsAndGrads();
      pg.Params := self.Biases.W;
      pg.Grads := self.Biases.DW;
      pg.l2_decay_mul := self.L2_Decay_Mul;
      pg.l1_decay_mul := self.L1_Decay_Mul;
      response.Add(pg);
      Result := response;
    end;
    
    public property InputDepth: Int32 read __InputDepth write __InputDepth;
    public property InputWidth: Int32 read __InputWidth write __InputWidth;
    public property InputHeight: Int32 read __InputHeight write __InputHeight;
  End;
Type SigmoidLayer=class(ILayer)
Private __OutputDepth: Int32;
Private __OutputHeight: Int32;
Private __OutputWidth: Int32;
Private __InputDepth: Int32;
Private __InputHeight: Int32;
Private __InputWidth: Int32;
Private __Input: Vol;
Private __Output: Vol;
public Property OutputDepth: Int32 Read __OutputDepth Write __OutputDepth;
public Property OutputHeight: Int32 Read __OutputHeight Write __OutputHeight;
public Property OutputWidth: Int32 Read __OutputWidth Write __OutputWidth;
public Property InputDepth: Int32 Read __InputDepth Write __InputDepth;
public Property InputHeight: Int32 Read __InputHeight Write __InputHeight;
public Property InputWidth: Int32 Read __InputWidth Write __InputWidth;
public Property Input: Vol Read __Input Write __Input;
public Property Output: Vol Read __Output Write __Output;
public Constructor Create(inputDepth: Int32; inputWidth: Int32; inputHeight: Int32);
Begin
self.OutputDepth := inputDepth;
self.OutputWidth := inputWidth;
self.OutputHeight := inputHeight;
self.InputHeight := inputHeight;
self.InputWidth := inputWidth;
self.InputDepth := inputDepth;
End;
public Function &Forward(V: Vol; is_training: Boolean): Vol;
Begin
self.input := V;
var V2 := V.CloneAndZero();
var N := V.W.Length;
var V2w := V2.W;
var Vw := V.W;
for var i := 0 to N-1 do
Begin
V2w[i] := 1.0 / (1.0 + Single(Math.Exp(-Vw[i])));
End;
self.Output := V2;
Result:=self.Output;
End;
public Procedure Backward();
Begin
var V := self.input;
var V2 := self.Output;
var N := V.W.Length;
V.DW := new Single[N];
for var i := 0 to N-1 do
Begin
var v2wi := V2.W[i];
V.DW[i] := v2wi * (1.0 - v2wi) * V2.DW[i];
End;
End;
public Function getParamsAndGrads(): List<ParamsAndGrads>;
Begin
var response := new List<ParamsAndGrads>();
Result := response;
End;
End;
type
  ReluLayer = class(ILayer)
    private __OutputDepth: Int32;
    private __OutputHeight: Int32;
    private __OutputWidth: Int32;
    private __InputDepth: Int32;
    private __InputHeight: Int32;
    private __InputWidth: Int32;
    public property OutputDepth: Int32 read __OutputDepth write __OutputDepth;
    public property OutputWidth: Int32 read __OutputWidth write __OutputWidth;
    public property OutputHeight: Int32 read __OutputHeight write __OutputHeight;
    public num_inputs: Int32;
    public input: Vol;
    private __Output: Vol;
    public property Output: Vol read __Output write __Output;
    public constructor Create(inputDepth: Int32; inputWidth: Int32; inputHeight: Int32);
    begin
      self.OutputDepth := inputDepth; 
      self.OutputWidth := inputWidth;
      self.OutputHeight := inputHeight; 
      self.InputHeight := inputHeight;
      self.InputWidth := inputWidth;
      self.InputDepth := inputDepth; 
    end;
    
    public function &Forward(V: Vol; is_training: Boolean): Vol;
    begin
      self.input := V;
      var V2 := V.Clone();
      var N := V.W.Length;
      var V2w := V2.W;
      for var i := 0 to N - 1 do
      begin
        if (V2w[i] < 0)
          Then V2w[i] := 0;
      end;
      self.Output := V2;
      Result := self.Output;
    end;
    
    public procedure Backward();
    begin
      var V := self.input;
      var V2 := self.Output;
      var N := V.W.Length;
      V.DW := new Single[N];
      for var i := 0 to N - 1 do
      begin
        if (V2.W[i] <= 0)
          Then V.DW[i] := 0
        else V.DW[i] := V2.DW[i];
      end;
    end;
    
    public function getParamsAndGrads(): List<ParamsAndGrads>;
    begin
      var response := new List<ParamsAndGrads>();
      Result := response;
    end;
    
    public property InputDepth: Int32 read __InputDepth write __InputDepth;
    public property InputWidth: Int32 read __InputWidth write __InputWidth;
    public property InputHeight: Int32 read __InputHeight write __InputHeight;
  End;

type
  InputLayer = class(ILayer)
    private __OutputDepth: Int32;
    private __OutputHeight: Int32;
    private __OutputWidth: Int32;
    private __InputDepth: Int32;
    private __InputHeight: Int32;
    private __InputWidth: Int32;
    public property OutputDepth: Int32 read __OutputDepth write __OutputDepth;
    public property OutputWidth: Int32 read __OutputWidth write __OutputWidth;
    public property OutputHeight: Int32 read __OutputHeight write __OutputHeight;
    public num_inputs: Int32;
    public input: Vol;
    private __Output: Vol;
    public property Output: Vol read __Output write __Output;
    public constructor Create();
    begin
    end;
    
    public function &Forward(V: Vol; is_training: Boolean): Vol;
    begin
      self.input := V;
      var A := new Vol(1, 1, self.OutputDepth, 0.0);
      self.Output := V;
      result := self.Output;
    end;
    
    public procedure Backward();
    begin
    end;
    
    public function getParamsAndGrads(): List<ParamsAndGrads>;
    begin
      var response := new List<ParamsAndGrads>();
      Result := response;
    end;
    
    public property InputDepth: Int32 read __InputDepth write __InputDepth;
    public property InputWidth: Int32 read __InputWidth write __InputWidth;
    public property InputHeight: Int32 read __InputHeight write __InputHeight;
  End;

type
  Network = class
    public Layers: List<ILayer>;
    public LossLayer: ILossLayer;
    public others: List<Vol> := new List<Vol>();
    public event ForwardLayer: EventHandler;
    public constructor Create();
    begin
      self.Layers := new List<ILayer>();
    end;
    
    public function &Forward(V: Vol; is_training: Boolean): Vol;
    begin
      var act := self.Layers[0].&Forward(V, is_training);
      for Var i := 1 to self.Layers.Count - 1 do
      begin
        act := self.Layers[i].&Forward(act, is_training);
        if (self.ForwardLayer <> nil)
        Then
        begin
          self.ForwardLayer(self.Layers[i], new EventArgs());
        end;
      end;
      act := self.LossLayer.&Forward(act, is_training);
      Result := act;
    end;
    
    public function Backward(y: Int32): Single;
    begin
      var loss := self.LossLayer.Backward(y);
      for Var i := Layers.Count - 1 DownTo 0 do
      begin
        Layers[i].Backward();
      end;
      Result := loss;
    end;
    
    public function Backward(y: array of Single): Single;
    begin
      var loss := (self.LossLayer as RegressionLayer).Backward(y);
      for Var i := Layers.Count - 1 DownTo 0 do
      begin
        Layers[i].Backward();
      end;
      Result := loss;
    end;
    
    public function Backward(y: ClassOutput): Single;
    begin
      var loss := (self.LossLayer as RegressionLayer).Backward(y);
      for Var i := Layers.Count - 1 DownTo 0 do
      begin
        Layers[i].Backward();
      end;
      Result := loss;
    end;
    
    public function GetParamsAndGrads(): List<ParamsAndGrads>;
    begin
      var response := new List<ParamsAndGrads>();
      for Var i := 0 to Layers.Count - 1 do
      begin
        response.AddRange(Layers[i].getParamsAndGrads());
      end;
      response.AddRange(LossLayer.getParamsAndGrads());
      foreach var item in others do
      begin
        var r := new ParamsAndGrads();
        r.Grads := item.DW;
        r.Params := item.W;
        r.l1_decay_mul := 0;
        r.l2_decay_mul := 0;
        response.Add(r);
      end;
      Result := response;
    end;
    
    public function GetPrediction(): Int32;
    begin
      var s := self.LossLayer;
      var p := s.Output.W;
      var maxv := p[0];
      var maxi := 0;
      for Var i := 1 to p.Length - 1 do
      begin
        if (p[i] > maxv)
        Then
        begin
          maxv := p[i];
          maxi := i;
        end;
      end;
      Result := maxi;
    end;
    
    public function GetTop5Prediction(): Dictionary<Int32, Single>;
    begin
      var res := new Dictionary<Int32, Single>(5);
      var s := self.LossLayer;
      var p := s.Output.W;  
      for Var j := 0 to 4 do
      begin
        var maxi := 0;
        var maxv := 0.0;
        for Var i := 0 to p.Length - 1 do
        begin
          if ((p[i] > maxv) and (not res.Keys.Contains(i)))
          Then
          begin
            maxv := p[i];
            maxi := i;
          end;
        end;
        res.Add(maxi, maxv);
      end;
      Result := res;
    end;
    
    public function GetPrediction(temp: Single): Int32;
    begin
      var s := self.LossLayer;
      var p := s.Output.W;
      var maxv := p[0];
      var maxi := 0;
      for Var i := 1 to p.Length - 1 do
      begin
        var px := p[i] / temp;
        if (p[i] > maxv)
        Then
        begin
          maxv := px;
          maxi := i;
        end;
      end;
      Result := maxi;
    end;
    
    public procedure Save(path: String);
    begin
      var fs := new FileStream(path, FileMode.OpenOrCreate, FileAccess.ReadWrite);
      //TODO
    end;
    
    public class function Load(path: String): Network;
    begin
      var fs := new FileStream(path, FileMode.OpenOrCreate, FileAccess.ReadWrite);
      //TODO
    end;
  End;

type
  TrainingMethod = (sgd, adam, adagrad, adadelta, windowgrad, nesterov);

type
  TrainingResult = class
    public l2_decay_loss: Single;
    public l1_decay_loss: Single;
    public cost_loss: Single;
    public softmax_loss: Single;
    public loss: Single;
  End;

type
  Trainer = class
    public Net: Network;
    public learning_rate: Single;
    public l1_decay: Single;
    public l2_decay: Single;
    public batch_size: Int32;
    public method: TrainingMethod;
    public momentum: Single;
    public ro: Single;
    public eps: Single;
    public beta1: Single;
    public beta2: Single;
    public k: Int32;
    public gsum: List<array of Single>;
    public xsum: List<array of Single>;
    public Regression: Boolean;
    public constructor Create();
    begin
      self.learning_rate := 0.01;
      self.l1_decay := 0.0;
      self.l2_decay := 0.0;
      self.batch_size := 51;
      self.method := TrainingMethod.sgd;
      self.momentum := 0.9;
      self.ro := 0.95;
      self.eps := 1e-8;
      self.beta1 := 0.9;
      self.beta2 := 0.999;
      self.k := 0;
      self.gsum := new List<array of Single>();
      self.xsum := new List<array of Single>();
    end;
    
    public function Train(x: Vol; y: Int32): TrainingResult;
    begin
      if (Net.LossLayer is RegressionLayer)
        Then self.Regression := true
      Else self.Regression := false;
      self.Net.&Forward(x, true);
      var cost_loss := self.Net.Backward(y);
      var l2_decay_loss := 0.0;
      var l1_decay_loss := 0.0;
      self.k += 1;
      if(self.k mod self.batch_size = 0)
      Then
      begin
        var pglist := self.Net.GetParamsAndGrads();
        if ((self.gsum.Count = 0) and ((self.method <> TrainingMethod.sgd) or (self.momentum > 0.0)))
        Then
        begin
          for Var i := 0 to pglist.Count - 1 do
          begin
            self.gsum.Add(new Single[pglist[i].Params.Length]);
            if ((self.method = TrainingMethod.adam) or (self.method = TrainingMethod.adadelta))
            Then
            begin
              self.xsum.Add(new Single[pglist[i].Params.Length]);
            end;
          end;
        end;
        for Var i := 0 to pglist.Count - 1 do
        begin
          var pg := pglist[i];
          var p := pg.Params;
          var g := pg.Grads;
          var l2_decay_mul := pg.l2_decay_mul;
          var l1_decay_mul := pg.l1_decay_mul;
          var l2_decay := self.l2_decay * l2_decay_mul;
          var l1_decay := self.l1_decay * l2_decay_mul;
          var plen := p.Length;
          for Var j := 0 to plen - 1 do
          begin
            l2_decay_loss += l2_decay * p[j] * p[j] / 2;
            l1_decay_loss += l1_decay * Math.Abs(p[j]);
            var l1grad := l1_decay * ((p[j] > 0) ? 1 : -1);
            var l2grad := l2_decay * p[j];
            var gij := (l2grad + l1grad + g[j]) / self.batch_size;
            var gsumi := self.gsum[i];
            if(self.method = TrainingMethod.adam)
            Then
            begin
              var xsumi := self.xsum[i];
              gsumi[j] := gsumi[j] * self.beta1 + (1 - self.beta1) * gij;
              xsumi[j] := xsumi[j] * self.beta2 + (1 - self.beta2) * gij * gij;
              var biasCorr1 := gsumi[j] * (1 - Math.Pow(self.beta1, self.k));
              var biasCorr2 := xsumi[j] * (1 - Math.Pow(self.beta2, self.k));
              var dx := -self.learning_rate * biasCorr1 / (Math.Sqrt(biasCorr2) + self.eps);
              p[j] += Single(dx);
            end
            else
            if(self.method = TrainingMethod.adagrad)
            Then
            begin
              gsumi[j] := gsumi[j] + gij * gij;
              var dx := -self.learning_rate / Math.Sqrt(gsumi[j] + self.eps) * gij;
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.windowgrad)
            Then
            begin
              gsumi[j] := self.ro * gsumi[j] + (1 - self.ro) * gij * gij;
              var dx := -self.learning_rate / Math.Sqrt(gsumi[j] + self.eps) * gij;
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.adadelta)
            Then
            begin
              var xsumi := self.xsum[i];
              gsumi[j] := self.ro * gsumi[j] + (1 - self.ro) * gij * gij;
              var dx := -Math.Sqrt((xsumi[j] + self.eps) / (gsumi[j] + self.eps)) * gij;
              xsumi[j] := self.ro * xsumi[j] + (1 - self.ro) * Single(dx) * Single(dx);
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.nesterov)
            Then
            begin
              var dx := gsumi[j];
              gsumi[j] := gsumi[j] * self.momentum + self.learning_rate * gij;
              dx := self.momentum * dx - (1.0 + self.momentum) * gsumi[j];
              p[j] += dx;
            end
            else
            begin
              if (self.momentum > 0.0)
              Then
              begin
                var dx := self.momentum * gsumi[j] - self.learning_rate * gij;
                gsumi[j] := dx;
                p[j] += dx;
              end
              else
              begin
                p[j] += -self.learning_rate * gij;
              end;
            end;
            g[j] := 0.0;
          end;
        end;
      end;
      var res := new TrainingResult();
      res.l2_decay_loss := l2_decay_loss;
      res.l1_decay_loss := l1_decay_loss;
      res.cost_loss := cost_loss;
      res.softmax_loss := cost_loss;
      res.loss := cost_loss + l1_decay_loss + l2_decay_loss;
      Result := res;
    end;
    
    public function Train(x: Vol; y: ClassOutput): TrainingResult;
    begin
      if (Net.LossLayer is RegressionLayer)
        Then self.Regression := true
      else self.Regression := false;
      self.Net.&Forward(x, true);
      var cost_loss := self.Net.Backward(y);
      var l2_decay_loss := 0.0;
      var l1_decay_loss := 0.0;
      self.k += 1;
      if(self.k mod self.batch_size = 0)
      Then
      begin
        var pglist := self.Net.GetParamsAndGrads();
        self.gsum.Clear();
        for Var i := 0 to pglist.Count - 1 do
        begin
          self.gsum.Add(new Single[pglist[i].Params.Length]);
        end;
        if ((self.gsum.Count = 0) and ((self.method <> TrainingMethod.sgd) or (self.momentum > 0.0)))
        Then
        begin
          for Var i := 0 to pglist.Count - 1 do
          begin
            if ((self.method = TrainingMethod.adam) or (self.method = TrainingMethod.adadelta))
            Then
            begin
              self.xsum.Add(new Single[pglist[i].Params.Length]);
            end;
          end;
        end;
        for Var i := 0 to pglist.Count - 1 do
        begin
          var pg := pglist[i];
          var p := pg.Params;
          var g := pg.Grads;
          var l2_decay_mul := pg.l2_decay_mul;
          var l1_decay_mul := pg.l1_decay_mul;
          var l2_decay := self.l2_decay * l2_decay_mul;
          var l1_decay := self.l1_decay * l2_decay_mul;
          var plen := p.Length;
          for Var j := 0 to plen - 1 do
          begin
            l2_decay_loss += l2_decay * p[j] * p[j] / 2;
            l1_decay_loss += l1_decay * Math.Abs(p[j]);
            var l1grad := l1_decay * ((p[j] > 0) ? 1 : -1);
            var l2grad := l2_decay * p[j];
            var gij := (l2grad + l1grad + g[j]) / self.batch_size;
            var gsumi := self.gsum[i];
            if(self.method = TrainingMethod.adam)
            Then
            begin
              var xsumi := self.xsum[i];
              gsumi[j] := gsumi[j] * self.beta1 + (1 - self.beta1) * gij;
              xsumi[j] := xsumi[j] * self.beta2 + (1 - self.beta2) * gij * gij;
              var biasCorr1 := gsumi[j] * (1 - Math.Pow(self.beta1, self.k));
              var biasCorr2 := xsumi[j] * (1 - Math.Pow(self.beta2, self.k));
              var dx := -self.learning_rate * biasCorr1 / (Math.Sqrt(biasCorr2) + self.eps);
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.adagrad)
            Then
            begin
              gsumi[j] := gsumi[j] + gij * gij;
              var dx := -self.learning_rate / Math.Sqrt(gsumi[j] + self.eps) * gij;
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.windowgrad)
            Then
            begin
              gsumi[j] := self.ro * gsumi[j] + (1 - self.ro) * gij * gij;
              var dx := -self.learning_rate / Math.Sqrt(gsumi[j] + self.eps) * gij;
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.adadelta)
            Then
            begin
              var xsumi := self.xsum[i];
              gsumi[j] := self.ro * gsumi[j] + (1 - self.ro) * gij * gij;
              var dx := -Math.Sqrt((xsumi[j] + self.eps) / (gsumi[j] + self.eps)) * gij;
              xsumi[j] := self.ro * xsumi[j] + (1 - self.ro) * Single(dx) * Single(dx);
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.nesterov)
            Then
            begin
              var dx := gsumi[j];
              gsumi[j] := gsumi[j] * self.momentum + self.learning_rate * gij;
              dx := self.momentum * dx - (1.0 + self.momentum) * gsumi[j];
              p[j] += dx;
            end
            else
            begin
              if (self.momentum > 0.0)
              Then
              begin
                var dx := self.momentum * gsumi[j] - self.learning_rate * gij;
                gsumi[j] := dx;
                p[j] += dx;
              end
              else
              begin
                p[j] += -self.learning_rate * gij;
              end;
            end;
            g[j] := 0.0;
          end;
        end;
      end;
      var res := new TrainingResult();
      res.l2_decay_loss := Single(l2_decay_loss);
      res.l1_decay_loss := Single(l1_decay_loss); 
      res.cost_loss := cost_loss;
      res.softmax_loss := cost_loss;
      res.loss := Single(cost_loss) + Single(l1_decay_loss) + Single(l2_decay_loss);
      Result := res;
    end;
    
    public function Train(x: Vol; y: array of Single): TrainingResult;
    begin
      if (Net.LossLayer is RegressionLayer)
        Then self.Regression := true
      else self.Regression := false;
      self.Net.&Forward(x, true);
      var cost_loss := self.Net.Backward(y);
      var l2_decay_loss := 0.0;
      var l1_decay_loss := 0.0;
      self.k += 1;
      if (self.k mod self.batch_size = 0)
      Then
      begin
        var pglist := self.Net.GetParamsAndGrads();
        if ((self.gsum.Count = 0) and ((self.method <> TrainingMethod.sgd) or (self.momentum > 0.0)))
        Then
        begin
          for Var i := 0 to pglist.Count - 1 do
          begin
            self.gsum.Add(new Single[pglist[i].Params.Length]);
            if ((self.method = TrainingMethod.adam) or (self.method = TrainingMethod.adadelta))
              Then self.xsum.Add(new Single[pglist[i].Params.Length]);
          end;
        end;
        for Var i := 0 to pglist.Count - 1 do
        begin
          var pg := pglist[i];
          var p := pg.Params;
          var g := pg.Grads;
          var l2_decay_mul := pg.l2_decay_mul;
          var l1_decay_mul := pg.l1_decay_mul;
          var l2_decay := self.l2_decay * l2_decay_mul;
          var l1_decay := self.l1_decay * l2_decay_mul;
          var plen := p.Length;
          for Var j := 0 to plen - 1 do
          begin
            l2_decay_loss += l2_decay * p[j] * p[j] / 2;
            l1_decay_loss += l1_decay * Math.Abs(p[j]);
            var l1grad := l1_decay * ((p[j] > 0) ? 1 : -1);
            var l2grad := l2_decay * (p[j]);
            var gij := (l2grad + l1grad + g[j]) / self.batch_size;
            var gsumi := self.gsum[i];
            if (self.method = TrainingMethod.adam)
            Then
            begin
              var xsumi := self.xsum[i];
              gsumi[j] := gsumi[j] * self.beta1 + (1 - self.beta1) * gij;
              xsumi[j] := xsumi[j] * self.beta2 + (1 - self.beta2) * gij * gij;
              var biasCorr1 := gsumi[j] * (1 - Math.Pow(self.beta1, self.k));
              var biasCorr2 := xsumi[j] * (1 - Math.Pow(self.beta2, self.k));
              var dx := -self.learning_rate * biasCorr1 / (Math.Sqrt(biasCorr2) + self.eps);
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.adagrad)
            Then
            begin
              gsumi[j] := gsumi[j] + gij * gij;
              var dx := -self.learning_rate / Math.Sqrt(gsumi[j] + self.eps) * gij;
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.windowgrad)
            Then
            begin
              gsumi[j] := self.ro * gsumi[j] + (1 - self.ro) * gij * gij;
              var dx := -self.learning_rate / Math.Sqrt(gsumi[j] + self.eps) * gij;
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.adadelta)
            Then
            begin
              var xsumi := self.xsum[i];
              gsumi[j] := self.ro * gsumi[j] + (1 - self.ro) * gij * gij;
              var dx := -Math.Sqrt((xsumi[j] + self.eps) / (gsumi[j] + self.eps)) * gij;
              xsumi[j] := self.ro * xsumi[j] + (1 - self.ro) * Single(dx) * Single(dx);
              p[j] += Single(dx);
            end
            else
            if (self.method = TrainingMethod.nesterov)
            Then
            begin
              var dx := gsumi[j];
              gsumi[j] := gsumi[j] * self.momentum + self.learning_rate * gij;
              dx := self.momentum * dx - (1.0 + self.momentum) * gsumi[j];
              p[j] += dx;
            end
            else
            begin
              if (self.momentum > 0.0)
              Then
              begin
                var dx := self.momentum * gsumi[j] - self.learning_rate * gij;
                gsumi[j] := dx;
                p[j] += dx;
              end
              else
              begin
                p[j] += -self.learning_rate * gij;
              end;
            end;
            g[j] := 0.0;
          end;
        end;
      end;
      var res := new TrainingResult();
      res.l2_decay_loss := Single(l2_decay_loss);
      res.l1_decay_loss := Single(l1_decay_loss);
      res.cost_loss := cost_loss;
      res.softmax_loss := cost_loss;
      res.loss := Single(cost_loss) + Single(l1_decay_loss) + Single(l2_decay_loss);
      Result := res;
    end;
  End;

implementation
End.